/*

First attempt at an Intel 8080 CPU emulator. Well-tested by
now, bugs seem to be eliminated. :-)

*/

#include "8080.h"
#include "io.h"

#include <stdint.h>

#if DEBUG
#include <stdio.h>
#endif

#define Bit6to7(x) (((x) >> 6) & 0x03)
#define Bit4to5(x) (((x) >> 4) & 0x03)
#define Bit3to5(x) (((x) >> 3) & 0x07)
#define Bit0to2(x) ((x) & 0x07)
#define Hide3to5(x) ((x) & 0xc7)
#define Hide4to5(x) ((x) & 0xcf)
#define Hide0to2(x) ((x) & 0xf8)

#define GetWord(x, y) (((word) (x) << 8) + (y))

typedef union reg_pair {
	uint8_t x[2];
	uint16_t y;
} reg_pair;

reg_pair reg[6];

#define A (reg[0].x[1])
#define F (reg[0].x[0]) /* Also called PSW */
#define B (reg[1].x[1])
#define C (reg[1].x[0])
#define D (reg[2].x[1])
#define E (reg[2].x[0])
#define H (reg[3].x[1])
#define L (reg[3].x[0])
#define SPH (reg[4].x[1])
#define SPL (reg[4].x[0])
#define PCH (reg[5].x[1])
#define PCL (reg[5].x[0])

#define AF (reg[0].y)
#define BC (reg[1].y)
#define DE (reg[2].y)
#define HL (reg[3].y)
#define SP (reg[4].y)
#define PC (reg[5].y)

#define GetCY ((F & 0x01) ? 1 : 0)
#define GetP ((F & 0x04) ? 1 : 0)
#define GetAC ((F & 0x10) ? 1 : 0)
#define GetZ ((F & 0x40) ? 1 : 0)
#define GetS ((F & 0x80) ? 1 : 0)

#define SetCY(n) (F = (F & ~0x01) | ((n) ? 0x01 : 0))
#define SetP(n) (F = (F & ~0x04) | ((n) ? 0x04 : 0))
#define SetAC(n) (F = (F & ~0x10) | ((n) ? 0x10 : 0))
#define SetZ(n) (F = (F & ~0x40) | ((n) ? 0x40 : 0))
#define SetS(n) (F = (F & ~0x80) | ((n) ? 0x80 : 0))

byte *reg8[8] = {&B, &C, &D, &E, &H, &L, 0, &A};
word *reg16[4] = {&BC, &DE, &HL, &SP};

extern int read_mem(word address);
extern int write_mem(word address, byte value);
extern int input(byte port); 
extern int output(byte port, byte value);
extern void resetINTE(void);
extern void setINTE(void);

static byte calc_parity(byte x);
static void setZSP(byte x);
static byte add(byte x, byte y, byte carry);
static byte sub(byte x, byte y, byte carry);
static void cmp(byte x, byte y);
static byte and(byte x, byte y);
static byte or(byte x, byte y);
static byte xor(byte x, byte y);
static byte inc(byte x);
static byte dec(byte x);
static word add16(word x, word y);
static word inc16(word x);
static word dec16(word x);
static void call(word x);
static void jump(word x);
static void ret(void);
static byte get_condition(byte x);

static byte calc_parity(byte x) {
	/* Calc parity (see Henry S. Warren "Hackers Delight", page 74) */
	byte y = x ^ (x >> 1);
	y ^= y >> 2;
	y ^= y >> 4;
	return !(y & 1);
}

byte parity[256];

static void setZSP(byte x) {
	SetZ(x == 0);
	SetS(x & 0x80);
	SetP(parity[x]);
}

static byte add(byte x, byte y, byte carry) {
	word sum = (word) x + y + carry;
	byte xsum = (x & 0xf) + (y & 0xf) + carry;
	byte zsum = sum & 0xff;

	setZSP(zsum);
	SetCY(zsum != sum);
	SetAC((xsum >> 4) & 1);

	return zsum;
}

static byte sub(byte x, byte y, byte borrow) {
	byte dif = add(x, ~y, !borrow);
	SetCY(!GetCY);
	return dif;
}

static void cmp(byte x, byte y) {
	sub(x, y, 0);
	return;
}

static byte and(byte x, byte y) {
#if AMD_CLONE
	SetAC(0);
#elif I8085
	SetAC(1);
#else
	SetAC(((x | y) >> 3) & 1);
#endif
	x &= y;
	setZSP(x);
	SetCY(0);

	return x;
}

static byte or(byte x, byte y) {
	x |= y;
	setZSP(x);
	SetCY(0);
	SetAC(0);

	return x;
}

static byte xor(byte x, byte y) {
	x ^= y;
	setZSP(x);
	SetCY(0);
	SetAC(0);

	return x;
}

static byte inc(byte x) {
	setZSP(++x);
	SetAC((x & 0x0f) == 0);
	return x;
}

static byte dec(byte x) {
	SetAC((((x & 0x0f) + 0xf) >> 4) & 1);
	setZSP(--x);
	return x;
}

static word add16(word x, word y) {
	uint32_t sum = (uint32_t) x + y;
	word zsum = (sum & 0xffff);
	SetCY(sum != zsum);
	return zsum;
}

static word inc16(word x) {
	return x + 1;
}

static word dec16(word x) {
	return x - 1;
}

static void jump(word x) {
	PC = x;
}

static void call(word x) {
	PC += 2;
	write_mem(SP - 1, PCH);
	write_mem(SP - 2, PCL);
	SP -= 2;
	PC = x;
}

static void ret(void) {
	PC = GetWord(read_mem(SP + 1), read_mem(SP));
	SP += 2;
}

static byte get_condition(byte x) {
	switch(x) {
		case 0: return !GetZ;
		case 1: return GetZ;
		case 2: return !GetCY;
		case 3: return GetCY;
		case 4: return !GetP;
		case 5: return GetP;
		case 6: return !GetS;
		case 7: return GetS;
	}
	return 0;
}

int init_emulator(word start) {
	int i;

	PC = start;
	SetP(1);

	for(i = 0; i < 256; i++) {
		parity[i] = calc_parity(i);
	}

	return 0;
}

unsigned long long run_emulator(unsigned long long cycles_max) {
	static unsigned long long cycles;
#if DEBUG
	static int ticker;
#endif
	int errorcode = 0;
	byte instr;

	word tempword;
	byte tempbyte;

	byte src, dest;

	/* Beginning of main loop... */
	while((!cycles_max || cycles < cycles_max) && !errorcode) {
		if((instr = read_mem(PC++)) < 0) {
			errorcode = READING_ERROR;
			break;
		}
#if DEBUG
		ticker++;
		if(ticker % DEBUG == 0)
			printf("PC: %04x, Instr: %02x A %02x B %02x C %02x D %02x "
				"E %02x H %02x L %02x SP %04x Z %1x CY %1x P %1x AC "
				"%1x S %1x\n", PC-1, instr, A, B, C, D, E, H, L, SP,
				GetZ, GetCY, GetP, GetAC, GetS);
#endif
		src = Bit0to2(instr);
		dest = Bit3to5(instr);

		if(Bit6to7(instr) == 1) {
			if(src != 6) {
				if(dest != 6) {
					/* MOV r1, r2 */
					*(reg8[dest]) = *(reg8[src]);
#if I8085
					cycles += 4;
#else
					cycles += 5;
#endif
				} else {
					/* MOV M, r */
					write_mem(HL, *(reg8[src]));
					cycles += 7;
				}
			} else {
				if(dest != 6) { /* 01...110 */
					/* MOV r, M */
					*(reg8[dest]) = read_mem(HL);
					cycles += 7;
				} else { /* 01110110 */
					/* HLT */
					errorcode = HALTED;
#if I8085
					cycles += 5;
#else
					cycles += 7;
#endif
				}
			}
		} else if(Hide3to5(instr) == 0x04) {
			if(dest != 6) {
				/* INR r */
				*(reg8[dest]) = inc(*(reg8[dest]));
#if I8085
				cycles += 4;
#else
				cycles += 5;
#endif
			} else {
				/* INR M */
				write_mem(HL, inc(read_mem(HL)));
				cycles += 10;
			} 
		} else if(Hide3to5(instr) == 0x05) {
			if(dest != 6) {
				/* DCR r */
				*(reg8[dest]) = dec(*(reg8[dest]));
#if I8085
				cycles += 4;
#else
				cycles += 5;
#endif
			} else {
				/* DCR M */
				write_mem(HL, dec(read_mem(HL)));
				cycles += 10;
			}
		} else if(Hide3to5(instr) == 0x06) {
			if(dest != 6) {
				/* MVI r, data */
				*(reg8[dest]) = read_mem(PC++);
				cycles += 7;
			} else {
				/* MVI M, data */
				write_mem(HL, read_mem(PC++));
				cycles += 10;
			}
		} else if(Hide3to5(instr) == 0xC0) {
			/* Rcondition */
			if(get_condition(dest)) {
				ret();
#if I8085
				cycles += 12;
#else
				cycles += 11;
#endif
			} else {
#if I8085
				cycles += 6;
#else
				cycles += 5;
#endif
			}
		} else if(Hide3to5(instr) == 0xC2) {
			/* Jcondition addr */
			if(get_condition(dest)) {
				jump(GetWord(read_mem(PC+1), read_mem(PC)));
				cycles += 10;
			} else {
				PC += 2;
#if I8085
				cycles += 7;
#else
				cycles += 10;
#endif
			}
		} else if(Hide3to5(instr) == 0xC4) {
			/* Ccondition addr */
			if(get_condition(dest)) {
				call(GetWord(read_mem(PC+1), read_mem(PC)));
#if I8085
				cycles += 18;
#else
				cycles += 17;
#endif
			} else {
				PC += 2;
#if I8085
				cycles += 9;
#else
				cycles += 11;
#endif
			}
		} else if(Hide3to5(instr) == 0xC7) {
			/* RST n */
			write_mem(SP - 1, PCH);
			write_mem(SP - 2, PCL);
			SP -= 2;
			PC = dest * 8;
#if I8085
			cycles += 12;
#else
			cycles += 11;
#endif
		} else if(Hide0to2(instr) == 0x80) {
			if(src != 6) {
				/* ADD r */
				A = add(A, *(reg8[src]), 0);
				cycles += 4;
			} else {
				/* ADD M */
				A = add(A, read_mem(HL), 0);
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0x88) {
			if(src != 6) {
				/* ADC r */
				A = add(A, *(reg8[src]), GetCY);
				cycles += 4;
			} else {
				/* ADC M */
				A = add(A, read_mem(HL), GetCY);
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0x90) {
			if(src != 6) {
				/* SUB r */
				A = sub(A, *(reg8[src]), 0);
				cycles += 4;
			} else {
				/* SUB M */
				A = sub(A, read_mem(HL), 0);
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0x98) {
			if(src != 6) {
				/* SBB r */
				A = sub(A, *(reg8[src]), GetCY);
				cycles += 4;
			} else {
				/* SBB M */
				A = sub(A, read_mem(HL), GetCY);
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0xA0) {
			if(src != 6) {
				/* ANA r */
				A = and(A, *(reg8[src]));
				cycles += 4;
			} else {
				/* ANA M */
				A = and(A, read_mem(HL));
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0xA8) {
			if(src != 6) {
				/* XRA r */
				A = xor(A, *(reg8[src]));
				cycles += 4;
			} else {
				/* XRA M */
				A = xor(A, read_mem(HL));
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0xB0) {
			if(src != 6) {
				/* ORA r */
				A = or(A, *(reg8[src]));
				cycles += 4;
			} else {
				/* ORA M */
				A = or(A, read_mem(HL));
				cycles += 7;
			}
		} else if(Hide0to2(instr) == 0xB8) {
			if (src != 6) {
				/* CMP r */
				cmp(A, *(reg8[src]));
				cycles += 4;
			} else {
				/* CMP M */
				cmp(A, read_mem(HL));
				cycles += 7;
			}
		} else if(Hide4to5(instr) == 0x01) {
			/* LXI rp, data16 */
			*(reg16[Bit4to5(instr)]) = GetWord(read_mem(PC+1),
				read_mem(PC));
			PC += 2;
			cycles += 10;
		} else if(Hide4to5(instr) == 0x03) {
			/* INX rp */
			*(reg16[Bit4to5(instr)]) = inc16(*(reg16[Bit4to5(instr)]));
#if I8085
			cycles += 6;
#else
			cycles += 5;
#endif
		} else if(Hide4to5(instr) == 0x09) {
			/* DAD rp */
			HL = (add16(HL, *(reg16[Bit4to5(instr)])));
			cycles += 10;
		} else if(Hide4to5(instr) == 0x0B) {
			/* DCX rp */
			*(reg16[Bit4to5(instr)]) = dec16(*(reg16[Bit4to5(instr)]));
#if I8085
				cycles += 6;
#else
				cycles += 5;
#endif
		} else if(Hide4to5(instr) == 0xC1) {
			if(Bit4to5(instr) != 3) {
				/* POP rp */
				*(reg16[Bit4to5(instr)]) = GetWord(read_mem(SP + 1),
					read_mem(SP));
			} else {
				/* POP PSW */
				A = read_mem(SP + 1);
				F = read_mem(SP);
			}
			SP += 2;
			cycles += 10;
		} else if(Hide4to5(instr) == 0xC5) {
			if(Bit4to5(instr) != 3) {
				/* PUSH rp */
				tempword = *(reg16[Bit4to5(instr)]);
				write_mem(SP - 1, tempword >> 8);
				write_mem(SP - 2, tempword & 0xff);
#if I8085
				cycles += 13;
#else
				cycles += 11;
#endif
			} else {
				/* PUSH PSW */
				write_mem(SP - 1, A);
				write_mem(SP - 2, (F & 0xd7) | 0x02);
#if I8085
				cycles += 12;
#else
				cycles += 11;
#endif
			}
			SP -= 2;
		} else switch(instr) {
			case 0x00:
				/* NOP */
				cycles += 4;
				break;
			case 0x02:
				/* STAX BC */
				write_mem(BC, A);
				cycles += 7;
				break;
			case 0x07:
				/* RLC */
				SetCY(A & 0x80);
				A = (A << 1) + GetCY;
				cycles += 4;
				break;
			case 0x0a:
				/* LDAX BC */
				A = read_mem(BC);
				cycles += 2;
				break;
			case 0x0f:
				/* RRC */
				SetCY(A & 0x01);
				A = (A >> 1) + (0x80 * GetCY);
				cycles += 4;
				break;
			case 0x12:
				/* STAX DE */
				write_mem(DE, A);
				cycles += 2;
				break;
			case 0x17:
				/* RAL */
				tempbyte = GetCY;
				SetCY(A & 0x80);
				A = (A << 1) + tempbyte;
				cycles += 4;
				break;
			case 0x1a:
				/* LDAX DE */
				A = read_mem(DE);
				cycles += 2;
				break;
			case 0x1f:
				/* RAR */
				tempbyte = GetCY;
				SetCY(A & 0x01);
				A = (A >> 1) + (0x80 * tempbyte);
				cycles += 4;
				break;
#if I8085
			case 0x20:
				/* RIM */
				/* TODO */
				cycles += 4;
#endif
			case 0x22:
				/* SHLD addr */
				tempword = GetWord(read_mem(PC+1), read_mem(PC));
				write_mem(tempword++, L);
				write_mem(tempword, H);
				PC += 2;
				cycles += 16;
				break;
			case 0x27:
				/* DAA */
				tempbyte = 0;
				if(GetAC || ((A & 0x0f) > 0x09)) {
					tempbyte = 0x06;
				}
				if((((A >> 4) > 9) || GetCY) ||
					(((A >> 4) == 9) && (GetCY || ((A & 0xf) > 9)))) {
					tempbyte |= 0x60;
				}
				SetAC((A & 0xf) >= 0xa);
				if((((A >> 4) >= 9) && ((A & 0xf) >= 0xa)) ||
					((A >> 4) >= 0xa)) {
					SetCY(1);
				}
				A += tempbyte;
				setZSP(A);
				cycles += 4;
				break;
			case 0x2a:
				/* LHLD addr */
				tempword = GetWord(read_mem(PC+1), read_mem(PC));
				L = read_mem(tempword++);
				H = read_mem(tempword);
				PC += 2;
				cycles += 16;
				break;
			case 0x2f:
				/* CMA */
				A = ~A;
				cycles += 4;
				break;
#if I8085
			case 0x30:
				/* SIM */
				cycles += 4;
				break;
#endif
			case 0x32:
				/* STA addr */
				write_mem(GetWord(read_mem(PC+1), read_mem(PC)), A);
				PC += 2;
				cycles += 13;
				break;
			case 0x37:
				/* STC */
				SetCY(1);
				cycles += 4;
				break;
			case 0x3a:
				/* LDA addr */
				A = read_mem(GetWord(read_mem(PC+1), read_mem(PC)));
				PC += 2;
				cycles += 13;
				break;
			case 0x3f:
				/* CMC */
				SetCY(!GetCY);
				cycles += 4;
				break;
			case 0xc3:
			case 0xcb:
				/* JMP addr */
				jump(GetWord(read_mem(PC+1), read_mem(PC)));
				cycles += 10;
				break;
			case 0xc6:
				/* ADI data */
				A = add(A, read_mem(PC++), 0);
				cycles += 7;
				break;
			case 0xc9:
			case 0xd9:
				/* RET */
				ret();
				cycles += 10;
				break;
			case 0xcd:
			case 0xdd:
			case 0xed:
			case 0xfd:
				/* CALL addr */
				call(GetWord(read_mem(PC+1), read_mem(PC)));
#if I8085
				cycles += 18;
#else
				cycles += 17;
#endif
				break;
			case 0xce:
				/* ACI data */
				A = add(A, read_mem(PC++), GetCY);
				cycles += 7;
				break;
			case 0xd3:
				/* OUT port */
				output(read_mem(PC++), A);
				cycles += 10;
				break;
			case 0xd6:
				/* SUI data */
				A = sub(A, read_mem(PC++), 0);
				cycles += 7;
				break;
			case 0xdb:
				/* IN port */
				tempword = input(read_mem(PC++));
				if(tempword == 1000) {
					return 1000;
				}
				A = tempword;
				cycles += 10;
				break;
			case 0xde:
				/* SBI data */
				A = sub(A, read_mem(PC++), GetCY);
				cycles += 7;
				break;
			case 0xe3:
				/* XTHL */
				tempword = GetWord(read_mem(SP+1), read_mem(SP));
				write_mem(SP, L);
				write_mem(SP+1, H);
				HL = tempword;
#if I8085
				cycles += 16;
#else
				cycles += 18;
#endif
				break;
			case 0xe6:
				/* ANI data */
				A = and(A, read_mem(PC++));
				cycles += 7;
				break;
			case 0xe9:
				/* PCHL */
				PC = HL;
#if I8085
				cycles += 6;
#else
				cycles += 5;
#endif
				break;
			case 0xeb:
				/* XCHG */
				tempword = HL;
				HL = DE;
				DE = tempword;
				cycles += 4;
				break;
			case 0xee:
				/* XRI data */
				A = xor(A, read_mem(PC++));
				cycles += 7;
				break;
			case 0xf3:
				/* DI */
				resetINTE();
				cycles += 4;
				break;
			case 0xf6:
				/* ORI data */
				A = or(A, read_mem(PC++));
				cycles += 7;
				break;
			case 0xf9:
				/* SPHL */
				SP = HL;
#if I8085
				cycles += 6;
#else
				cycles += 5;
#endif
				break;
			case 0xfb:
				/* EI */
				setINTE();
				cycles += 4;
				break;
			case 0xfe:
				/* CPI data */
				cmp(A, read_mem(PC++));
				cycles += 7;
				break;
			default:
				cycles += 4;
		}
	}
	/* End of main loop */

	cycles -= cycles_max;

	return errorcode;
}

